@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2010-2013 RT-RK
+   Copyright (C) 2010-2015 RT-RK
       mips-valgrind@rt-rk.com
 
    This program is free software; you can redistribute it and/or
@@ -71,12 +71,12 @@ typedef unsigned long long __vki_u64;
 // From linux-2.6.35.5/include/asm-mips/page.h
 //----------------------------------------------------------------------
 
-/* PAGE_SHIFT determines the page size */
-#define VKI_PAGE_SHIFT          MIPS_PAGE_SHIFT
-#define VKI_PAGE_SIZE           (1UL << VKI_PAGE_SHIFT)
+/* MIPS32 uses runtime pagesize detection */
+extern UWord VKI_PAGE_SHIFT;
+extern UWord VKI_PAGE_SIZE;
 #define VKI_PAGE_MASK           (~(VKI_PAGE_SIZE-1))
-#define VKI_MAX_PAGE_SHIFT      VKI_PAGE_SHIFT
-#define VKI_MAX_PAGE_SIZE       VKI_PAGE_SIZE
+#define VKI_MAX_PAGE_SHIFT      16
+#define VKI_MAX_PAGE_SIZE       (1UL << VKI_MAX_PAGE_SHIFT)
 
 //----------------------------------------------------------------------
 // From linux-2.6.35.9/arch/mips/include/bits/shm.h
@@ -343,6 +343,10 @@ struct vki_sigcontext {
 #define VKI_F_SETOWN_EX		15
 #define VKI_F_GETOWN_EX		16
 
+#define VKI_F_OFD_GETLK		36
+#define VKI_F_OFD_SETLK		37
+#define VKI_F_OFD_SETLKW	38
+
 #define VKI_F_GETLK64		33			/*  using 'struct flock64' */
 #define VKI_F_SETLK64		34
 #define VKI_F_SETLKW64		35
@@ -382,6 +386,7 @@ struct vki_f_owner_ex {
 
 #define VKI_SIOCSPGRP           0x8902
 #define VKI_SIOCGPGRP           0x8904
+#define VKI_SIOCATMARK          0x8905
 #define VKI_SIOCGSTAMP          0x8906      /* Get stamp (timeval) */
 #define VKI_SIOCGSTAMPNS        0x8907      /* Get stamp (timespec) */
 
